# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from typing import Any, Callable

from loguru import logger
from pydantic import BaseModel, field_validator

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class InnerTaskProfile(BaseModel):

    inner_task_name: str
    number_of_agents: int
    number_of_layers: int
    observation_space_sizes: dict[str, int]
    action_space_sizes: dict[str, int]
    number_of_parameters: int
    vram_usage: float
    idle_vram_usage: float

    @field_validator("vram_usage", "idle_vram_usage", mode="before")
    def replace_non_with_nan(cls, value: None | float) -> float:
        """
        :param value: Value to replace with NaN if it is None.
        :return: Either the given float value or NaN.
        """

        if value is None:
            return float("nan")

        return value

    @property
    def failed_to_profile(self) -> bool:
        """
        :return: Whether the corresponding task is completely invalid since it could not be profiled.
        """

        return math.isnan(self.vram_usage)

    def model_dump_json(self, **kwargs) -> str:
        """
        :param kwargs: Standard Pydantic model dump kwargs.
        :return: Dump result of the superclass' method.
        """

        logger.debug(
            f"Inner task {self.inner_task_name} consumed {self.vram_usage:.3f} MB of VRAM while training and "
            f"{self.idle_vram_usage:.3f} MB of VRAM while idle. It has {self.number_of_agents} agents across "
            f"{self.number_of_layers} inner model layers."
        )

        return super().model_dump_json(**kwargs)


class InnerTaskProfiles(BaseModel):

    profiles: dict[str, InnerTaskProfile] = {}
    rejected_tasks: list[str] = []

    @property
    def max_agents(self) -> int:
        """
        :return: Maximum number of agents active in any one environment.
        """

        return max(profile.number_of_agents for profile in self.profiles.values())

    @property
    def max_layers(self) -> int:
        """
        :return: Maximum number of inner model layers in any one environment.
        """

        return max(profile.number_of_layers for profile in self.profiles.values())

    @property
    def min_layers(self) -> int:
        """
        :return: Minimum number of inner model layers in any one environment.
        """

        return min(profile.number_of_layers for profile in self.profiles.values())

    @property
    def possible_agents(self) -> list[str]:
        """
        :return: A list of names of the possible agents across inner tasks.
        """

        return list(self.compiled_observation_sizes.keys())

    @property
    def compiled_observation_sizes(self) -> dict[str, int]:
        """
        :return: The observation size of each possible agent.
        """

        spaces = {
            inner_task_name: inner_task_profile.observation_space_sizes
            for inner_task_name, inner_task_profile in self.profiles.items()
        }
        return self._compile_gym_space_sizes(spaces=spaces)

    @property
    def compiled_action_sizes(self) -> dict[str, int]:
        """
        :return: The action size of each possible agent.
        """

        spaces = {
            inner_task_name: inner_task_profile.action_space_sizes
            for inner_task_name, inner_task_profile in self.profiles.items()
        }
        return self._compile_gym_space_sizes(spaces=spaces)

    @property
    def total_observation_size(self) -> int:
        """
        :return: Total observation size summed across all agents.
        """

        return sum(self.compiled_observation_sizes.values())

    @property
    def total_action_size(self) -> int:
        """
        :return: Total action size summed across all agents.
        """

        return sum(self.compiled_action_sizes.values())

    @property
    def max_total_observation_size(self) -> int:
        """
        :return: The summed observation size of all agents with the task that has the most layers.
        """

        if not self.profiles:
            raise ValueError(
                "No profiles to calculate max total observation size. Ensure profiles have been "
                "added before executing the training loop"
            )

        largest_task_name = max(self.profiles, key=lambda k: self.profiles[k].number_of_layers)
        largest_task = self.profiles[largest_task_name]

        return sum(largest_task.observation_space_sizes.values())

    @property
    def max_total_action_size(self) -> int:
        """
        :return: The summed action size of all agents with the task that has the most layers.
        """

        if not self.profiles:
            raise ValueError(
                "No profiles to calculate max total action size. Ensure profiles have been "
                "added before executing the training loop"
            )

        largest_task_name = max(self.profiles, key=lambda k: self.profiles[k].number_of_layers)
        largest_task = self.profiles[largest_task_name]

        return sum(largest_task.action_space_sizes.values())

    @staticmethod
    def _compile_gym_space_sizes(spaces: dict[str, dict[str, int]]) -> dict[str, int]:
        """
        :param spaces: Dictionary mapping inner task names to a dictionary mapping agent IDs to observation/action
        space sizes.
        :return: Dictionary mapping agent IDs to observation/action space sizes.
        """

        compiled_spaces: dict[str, int] = {}

        for inner_task_spaces in spaces.values():
            for agent_id, agent_space in inner_task_spaces.items():
                if agent_id not in compiled_spaces:
                    compiled_spaces[agent_id] = agent_space

                elif agent_space != compiled_spaces[agent_id]:
                    raise ValueError(
                        f"Gym space size for agent {agent_id} differs between environments. "
                        f"Expected {compiled_spaces[agent_id]}, Got {agent_space}."
                    )

        return compiled_spaces

    @staticmethod
    def validate_size(
        policy_mapping_function: Callable[[str, Any, Any], str],
        size_dictionary: dict[str, int],
        expected_sizes: dict[str, int],
        task_name: str,
    ) -> dict[str, int]:
        """
        :param policy_mapping_function: Function which maps agent IDs to policy IDs.
        :param size_dictionary: Dictionary mapping agent IDs to observation/action space sizes.
        :param expected_sizes: Dictionary mapping policy IDs to expected observation/action space sizes.
        :param task_name: Name of the task being validated.
        :return: Updated expected_sizes dictionary.
        """

        for agent_id, size in size_dictionary.items():
            policy_id = policy_mapping_function(agent_id)  # type: ignore

            if policy_id not in expected_sizes:
                expected_sizes[policy_id] = size

            elif expected_sizes[policy_id] != size:
                raise ValueError(
                    f"The observation space size of agent {agent_id} for task {task_name} does not match the "
                    f"expected size {size}."
                )

        return expected_sizes

    def profile_exists(self, inner_task_name: str) -> bool:
        """
        :param inner_task_name: Name of the inner task to check whether the profile exists.
        :return: Whether the profile of the given inner task exists.
        """

        return inner_task_name in self.profiles

    def add_profile(
        self,
        inner_task_name: str,
        number_of_agents: int,
        number_of_layers: int,
        observation_space_sizes: dict[str, int],
        action_space_sizes: dict[str, int],
        number_of_parameters: int,
        vram_usage: float,
        idle_vram_usage: float,
    ) -> None:
        """
        :param inner_task_name: Name of the inner task to add a profile for.
        :param number_of_agents: Number of agents active in the inner task's environment.
        :param number_of_layers: Number of layers in the inner model.
        :param observation_space_sizes: Dictionary mapping agent IDs to their observation space sizes.
        :param action_space_sizes: Dictionary mapping agent IDs to their action space sizes.
        :param vram_usage: VRAM required to perform the inner task. Can be NaN if an OOM was encountered.
        :param idle_vram_usage: VRAM required for the inner task to sit loaded but not actively being trained. Can be
        NaN if an OOM was encountered.
        """

        self.profiles[inner_task_name] = InnerTaskProfile(
            inner_task_name=inner_task_name,
            number_of_agents=number_of_agents,
            number_of_layers=number_of_layers,
            observation_space_sizes=observation_space_sizes,
            action_space_sizes=action_space_sizes,
            number_of_parameters=number_of_parameters,
            vram_usage=vram_usage,
            idle_vram_usage=idle_vram_usage,
        )

    def validate_task_profiles(self, policy_mapping_function: Callable[[str, Any, Any], str]) -> None:
        """
        :param policy_mapping_function: Function which maps agent IDs to policy IDs.
        """

        expected_observation_sizes: dict[str, int] = {}
        expected_action_sizes: dict[str, int] = {}

        for task_name, task_profile in self.profiles.items():
            expected_observation_sizes = self.validate_size(
                policy_mapping_function=policy_mapping_function,
                size_dictionary=task_profile.observation_space_sizes,
                expected_sizes=expected_observation_sizes,
                task_name=task_name,
            )
            expected_action_sizes = self.validate_size(
                policy_mapping_function=policy_mapping_function,
                size_dictionary=task_profile.action_space_sizes,
                expected_sizes=expected_action_sizes,
                task_name=task_name,
            )

    def reject_task(self, task: str) -> None:
        """
        :param task: Task being rejected due to excessive VRAM consumption.
        """

        if task not in self.rejected_tasks:
            self.rejected_tasks.append(task)

    def get_valid_profiles(self, force_tasks: str | list[str] | None = None) -> dict[str, InnerTaskProfile]:
        """
        :param force_tasks: Specific task or list of tasks to always include in valid profiles or None.
        :return: Dictionary mapping inner task names to profiles of the tasks with all tasks that failed to be profiled
        or were rejected removed.
        """

        force_tasks = [force_tasks] if isinstance(force_tasks, str) else force_tasks
        force_tasks = [] if force_tasks is None else force_tasks

        return {
            task: profile
            for task, profile in self.profiles.items()
            if task in force_tasks or (not profile.failed_to_profile and task not in self.rejected_tasks)
        }

    def get_vram_usages_of_valid_tasks(self) -> dict[str, float]:
        """
        :return: Dictionary mapping inner task names to the VRAM usage that task.
        """

        return {task: profile.vram_usage for task, profile in self.get_valid_profiles().items()}



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from typing import Any

from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import (
    LHOPTBaseObserver,
    LHOPTCheckpointBaseObserver,
)
from libinephany.observations.observers.global_observers.constants import LHOPT_CONSTANTS
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TrainingLoss(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.training_loss

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ValidationLoss(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.validation_loss

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class LossRatio(GlobalObserver):

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        if observation_inputs.validation_loss == 0:
            return 0

        return observation_inputs.training_loss / observation_inputs.validation_loss

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class TrainingScore(GlobalObserver):

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.training_score

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ValidationScore(GlobalObserver):

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.validation_score

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class LHOPTTrainingLoss(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns three-dimensional observations: [is_nan, is_inf, cdf_feature] for training loss values.

    This observer use the CDF calculation from the paper and applies CDF transformation using the CDF mean and std.
    """

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> list[int | float]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: List of three features: [is_nan, is_inf, cdf_feature]
        """

        training_loss = observation_inputs.training_loss

        cdf_feature = self._compute_cdf_feature(training_loss)

        self._update_time()

        return [int(math.isnan(training_loss)), int(math.isinf(training_loss)), cdf_feature]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}

    def reset(self) -> None:
        """Reset the observer by clearing the time series."""
        self._time_series: list[tuple[float, float]] = []
        self._current_time: float = 0.0

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 3  # [is_nan, is_inf, cdf_feature]


class LHOPTValidationLoss(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns three-dimensional observations: [is_nan, is_inf, cdf_feature] for validation loss values.

    This observer use the CDF calculation from the paper and applies CDF transformation using the CDF mean and std.
    """

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> list[int | float]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistics models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: List of three features: [is_nan, is_inf, cdf_feature]
        """

        validation_loss = observation_inputs.validation_loss

        cdf_feature = self._compute_cdf_feature(validation_loss)

        self._update_time()

        return [int(math.isnan(validation_loss)), int(math.isinf(validation_loss)), cdf_feature]

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}

    def reset(self) -> None:
        """Reset the observer by clearing the time series."""
        self._time_series: list[tuple[float, float]] = []
        self._current_time: float = 0.0

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 3  # [is_nan, is_inf, cdf_feature]


class LHOPTLossRatio(LHOPTBaseObserver):
    """
    This is a global observer from the OpenAI paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns three-dimensional observations: [is_nan, tanh, cdf_feature] for loss ratio values.

    This observer computes the logarithm of the ratio between validation_score and training_score,
    providing three features:
    1. is_nan - whether the log ratio is NaN
    2. tanh(log_ratio) - bounded feature using hyperbolic tangent
    3. cdf_feature - CDF transformed feature using CDF mean and std
    """

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> list[int | float]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistics models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: List of three features: [is_nan, tanh, cdf_feature]
        """

        log_ratio = self._compute_log_ratio(
            training_score=observation_inputs.training_score, validation_score=observation_inputs.validation_score
        )

        tanh_feature = math.tanh(max(-LHOPT_CONSTANTS["TANH_BOUND"], min(LHOPT_CONSTANTS["TANH_BOUND"], log_ratio)))

        cdf_feature = self._compute_cdf_feature(log_ratio)

        self._update_time()

        return [int(math.isnan(log_ratio)), tanh_feature, cdf_feature]

    def _compute_log_ratio(self, training_score: float, validation_score: float) -> float:
        """
        Compute the log ratio of validation_score to training_score.

        :param training_score: Training score value
        :param validation_score: Validation score value
        :return: Log ratio value
        """
        if training_score <= 0:
            return 0.0

        if validation_score <= 0:
            return 0.0

        # Calculate the ratio of validation_score to training_score
        score_ratio = validation_score / training_score

        return math.log(score_ratio)

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """
        return 3  # [is_nan, tanh, cdf_feature]


class PercentileOfLossAtEachCheckpoint(LHOPTCheckpointBaseObserver):
    """
    Observer that computes percentile of loss values at each checkpoint.
    """

    def __init__(
        self,
        checkpoint_interval: int = LHOPT_CONSTANTS["DEFAULT_CHECKPOINT_INTERVAL"],
        percentile: float = LHOPT_CONSTANTS["DEFAULT_PERCENTILE"],
        **kwargs,
    ) -> None:
        """
        :param checkpoint_interval: How often to create checkpoints (in training steps).
        :param percentile: Percentile to compute (0.0 to 1.0).
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(checkpoint_interval=checkpoint_interval, **kwargs)
        self.percentile = max(0.0, min(1.0, percentile))

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Percentile value of loss at checkpoint.
        """
        training_loss = observation_inputs.training_loss

        # Handle cold start
        self._cold_start(training_loss)

        # Update history
        self._update_history(training_loss)

        # Check if we should create a checkpoint
        if self._should_create_checkpoint():
            # Compute percentile
            sorted_history = sorted(self._history)
            index = int(self.percentile * (len(sorted_history) - 1))
            percentile_value = sorted_history[index]

            self._cached_observation = percentile_value
            return percentile_value
        else:
            # Return last value during warm-up
            self._cached_observation = self.last_value
            return self.last_value



# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

from typing import TypedDict


class LHOPTConstants(TypedDict):
    IS_NAN: float
    NOT_NAN: float
    IS_INF: float
    NOT_INF: float
    TANH_BOUND: float
    DEFAULT_DECAY_FACTOR: float
    DEFAULT_TIME_WINDOW: int
    DEFAULT_CHECKPOINT_INTERVAL: int
    DEFAULT_PERCENTILE: float
    ZERO_DIVISION_TOLERANCE: float
    DEFAULT_SAMPLE_FREQUENCY: int
    DEFAULT_VARIANCE_THRESHOLD: float
    DEFAULT_ENV_STEP_SAMPLE_FREQUENCY: int


# Create the constants instance
LHOPT_CONSTANTS: LHOPTConstants = LHOPTConstants(
    IS_NAN=1.0,
    NOT_NAN=0.0,
    IS_INF=1.0,
    NOT_INF=0.0,
    TANH_BOUND=10.0,
    DEFAULT_DECAY_FACTOR=1.25,
    DEFAULT_TIME_WINDOW=32,
    DEFAULT_CHECKPOINT_INTERVAL=100,
    DEFAULT_PERCENTILE=0.6,
    ZERO_DIVISION_TOLERANCE=1e-8,
    DEFAULT_SAMPLE_FREQUENCY=4,
    DEFAULT_VARIANCE_THRESHOLD=1e-6,
    DEFAULT_ENV_STEP_SAMPLE_FREQUENCY=10,
)



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import torch
from pydantic import BaseModel

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

FIELD_SUFFIX = "_"

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TensorStatistics(BaseModel):

    norm_: float = 0.0
    mean_: float = 0.0
    median_: float = 0.0
    variance_: float = 0.0
    tenth_percentile_: float = 0.0
    ninetieth_percentile_: float = 0.0
    inter_quartile_range_: float = 0.0

    @property
    def norm(self) -> float:
        """
        :return: Tensor norm.
        """

        return self.norm_

    @norm.setter
    def norm(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the norm from.
        """

        self.norm_ = torch.norm(tensor, p=2).item()

    @property
    def mean(self) -> float:
        """
        :return: Tensor mean.
        """

        return self.mean_

    @mean.setter
    def mean(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the mean from.
        """

        self.mean_ = tensor.mean().item()

    @property
    def median(self) -> float:
        """
        :return: Tensor median.
        """

        return self.median_

    @median.setter
    def median(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the median from.
        """

        self.median_ = tensor.median().item()

    @property
    def variance(self) -> float:
        """
        :return: Tensor variance.
        """

        return self.variance_

    @variance.setter
    def variance(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the variance from.
        """

        self.variance_ = tensor.var(unbiased=False).item()

    @property
    def tenth_percentile(self) -> float:
        """
        :return: Tensor tenth percentile.
        """

        return self.tenth_percentile_

    @tenth_percentile.setter
    def tenth_percentile(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the tenth percentile from.
        """

        self.tenth_percentile_ = torch.quantile(tensor, 0.1).item()

    @property
    def ninetieth_percentile(self) -> float:
        """
        :return: Tensor ninetieth percentile.
        """

        return self.ninetieth_percentile_

    @ninetieth_percentile.setter
    def ninetieth_percentile(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the ninetieth from.
        """

        self.ninetieth_percentile_ = torch.quantile(tensor, 0.9).item()

    @property
    def inter_quartile_range(self) -> float:
        """
        :return: Tensor inter-quartile range.
        """

        return self.inter_quartile_range_

    @inter_quartile_range.setter
    def inter_quartile_range(self, tensor: torch.Tensor) -> None:
        """
        :param tensor: Tensor to compute the inter-quartile range from.
        """

        self.inter_quartile_range_ = (torch.quantile(tensor, 0.75) - torch.quantile(tensor, 0.25)).item()

    @staticmethod
    def downsample_tensor(tensor: torch.Tensor, sample_percentage: float) -> torch.Tensor:
        """
        :param tensor: Tensor to downsample.
        :param sample_percentage: Percentage of the given tensor to randomly sample and compute statistics from.
        :return: Downsampled tensor.
        """

        if sample_percentage >= 1.0:
            return tensor

        input_size = tensor.numel()
        sample_size = max(int(input_size * sample_percentage), 1)

        random_indices = torch.randint(0, input_size, (sample_size,), device=tensor.device)
        tensor = tensor.view(-1)

        return tensor[random_indices]

    @classmethod
    def filter_include_statistics(cls, include_statistics: list[str]) -> list[str]:
        """
        :param include_statistics: Names of the fields in the model to include in returned observations.
        :return: List of fields from the given include_statistics list that are present in this pydantic model.
        :raises ValueError: If no statistics to include are given.
        """

        filtered_include_statistics: list[str] = []

        for include_stat in include_statistics:
            with_suffix = include_stat + FIELD_SUFFIX if not include_stat.endswith(FIELD_SUFFIX) else include_stat

            if with_suffix in cls.model_fields.keys():
                filtered_include_statistics.append(with_suffix)

        if not filtered_include_statistics:
            raise ValueError(f"No statistics to include given to {cls.__name__}!")

        return filtered_include_statistics

    @classmethod
    def build(
        cls,
        tensor: torch.Tensor,
        include_statistics: list[str],
        sample_percentage: float = 0.01,
    ) -> "TensorStatistics":
        """
        :param tensor: Tensor to compute and store statistics of.
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations.
        :param sample_percentage: Percentage of the given tensor to randomly sample and compute statistics from.
        :return: Constructed tensor statistics.
        """

        stats = cls()
        downsampled_tensor = cls.downsample_tensor(tensor=tensor, sample_percentage=sample_percentage)

        for field, field_value in stats.model_dump().items():
            name = field[:-1] if field.endswith(FIELD_SUFFIX) else field

            if name in include_statistics:
                setattr(stats, name, downsampled_tensor)

        return stats

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor) -> "TensorStatistics":
        """
        :param tensor: Tensor to build the model from.
        :return: Reconstructed model.
        """

        return cls(
            norm_=tensor[0],
            mean_=tensor[1],
            median_=tensor[2],
            variance_=tensor[3],
            tenth_percentile_=tensor[4],
            ninetieth_percentile_=tensor[5],
            inter_quartile_range_=tensor[6],
        )

    def to_list(self, include_statistics: list[str]) -> list[float]:
        """
        :param include_statistics: List of field names to include in the returned list.
        :return: List of field values.
        """

        filtered_includes = self.filter_include_statistics(include_statistics=include_statistics)

        as_list = []

        for field, field_value in self.model_dump().items():
            without_suffix = field[:-1]

            if field in filtered_includes or without_suffix in filtered_includes:
                as_list.append(field_value)

        return as_list

    def to_tensor(self) -> torch.Tensor:
        """
        :return: Tensor with contents of the Pydantic model in a specific order.
        """

        return torch.tensor(
            [
                self.norm,
                self.mean,
                self.median,
                self.variance,
                self.tenth_percentile,
                self.ninetieth_percentile,
                self.inter_quartile_range,
            ]
        )

    def as_observation_dict(self) -> dict[str, float]:
        """
        :return: Dictionary of observation values.
        """

        return {
            field[:-1] if field.endswith(FIELD_SUFFIX) else field: field_value
            for field, field_value in self.model_dump().items()
        }



# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

KEY_HEADER_CASE = "X-API-Key"
KEY_HEADER_NO_CASE = KEY_HEADER_CASE.lower()

TIMESTAMP_FORMAT = "%Y-%m-%d-%H-%M-%S"
TIMESTAMP_FORMAT_WITH_MS = "%Y-%m-%d-%H-%M-%S-%f"

RLLIB_TRUNC_EPISODES = "truncate_episodes"
RLLIB_COMP_EPISODES = "complete_episodes"

LEARNING_RATE = "learning_rate"
WEIGHT_DECAY = "weight_decay"
DROPOUT = "dropout"
GRAD_NORM_CLIP = "grad_norm_clip"
ADAM_BETA_ONE = "adam_beta_one"
ADAM_BETA_TWO = "adam_beta_two"
ADAM_EPS = "adam_eps"
SGD_MOMENTUM = "sgd_momentum"
BATCH_SIZE = "batch_size"
GRADIENT_ACCUMULATION = "gradient_accumulation"
EPOCHS = "epochs"
TOKENS = "tokens"
SAMPLES = "samples"

PARAMS = "params"

SCHEDULER_GROUP_NAME = "inephany_parameter_group_name"

AGENT_PREFIX_LR = "lr"
AGENT_PREFIX_WD = "weight-decay"
AGENT_PREFIX_DROPOUT = "dropout"
AGENT_PREFIX_CLIPPING = "grad-norm-clip"

AGENT_PREFIX_BETA_ONE = "adam-beta-one"
AGENT_PREFIX_BETA_TWO = "adam-beta-two"
AGENT_PREFIX_EPS = "adam-eps"

AGENT_PREFIX_SGD_MOMENTUM = "sgd-momentum"

AGENT_BATCH_SIZE = "batch-size"
AGENT_PREFIX_GRADIENT_ACCUMULATION = "gradient-accumulation"

AGENT_BANDIT_SUFFIX = "bandit-agent"

AGENT_TYPES = [
    LEARNING_RATE,
    WEIGHT_DECAY,
    DROPOUT,
    GRAD_NORM_CLIP,
    ADAM_BETA_ONE,
    ADAM_BETA_TWO,
    ADAM_EPS,
    SGD_MOMENTUM,
    GRADIENT_ACCUMULATION,
]
SUFFIXES = [AGENT_BANDIT_SUFFIX]
PREFIXES = [
    AGENT_PREFIX_LR,
    AGENT_PREFIX_WD,
    AGENT_PREFIX_DROPOUT,
    AGENT_PREFIX_CLIPPING,
    AGENT_PREFIX_BETA_ONE,
    AGENT_PREFIX_BETA_TWO,
    AGENT_PREFIX_EPS,
    AGENT_PREFIX_SGD_MOMENTUM,
    AGENT_PREFIX_GRADIENT_ACCUMULATION,
]
PREFIXES_TO_HPARAMS = {
    AGENT_PREFIX_LR: LEARNING_RATE,
    AGENT_PREFIX_WD: WEIGHT_DECAY,
    AGENT_PREFIX_DROPOUT: DROPOUT,
    AGENT_PREFIX_CLIPPING: GRAD_NORM_CLIP,
    AGENT_PREFIX_BETA_ONE: ADAM_BETA_ONE,
    AGENT_PREFIX_BETA_TWO: ADAM_BETA_TWO,
    AGENT_PREFIX_EPS: ADAM_EPS,
    AGENT_PREFIX_SGD_MOMENTUM: SGD_MOMENTUM,
    AGENT_PREFIX_GRADIENT_ACCUMULATION: GRADIENT_ACCUMULATION,
}
HPARAMS_TO_PREFIXES = {hparam: prefix for prefix, hparam in PREFIXES_TO_HPARAMS.items()}



# ======================================================================================================================
#
# BACKWARD COMPATIBILITY MODULE
#
# This module provides backward compatibility for older pickled models that reference
# libinephany.pydantic_models.configs.outer_model_config. The actual implementation
# has moved to libipcode.pydantic_models.configs.outer_model_config.
#
# This allows older models to deserialize without moving the config back to libinephany.
#
# ======================================================================================================================

try:
    # Re-export all classes from libipcode to maintain backward compatibility
    from libipcode.pydantic_models.configs.outer_model_config import OuterModelConfig, SharedActorCriticConfig
except ImportError:
    # If libipcode is not available, provide a fallback or raise a more helpful error
    raise ImportError(
        "libipcode is required for backward compatibility with older models. Ignore this message if you"
        " are not an Inephany developer."
    )

# Export all symbols to maintain the same interface
__all__ = [
    "OuterModelConfig",
    "SharedActorCriticConfig",
]



# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================

from libinephany.utils.constants import PREFIXES, PREFIXES_TO_HPARAMS, SUFFIXES

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

ID_CONNECTOR = "-"

# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def _strip_substrings(strip_from: str, substrings: list[str]) -> str:
    """
    :param strip_from: String to strip given substrings from.
    :param substrings: Substrings to strip from the given string.
    :return: Given string stripped of given substrings.
    """

    for substring in substrings:
        strip_from = strip_from.replace(substring, "")

    if strip_from[0] == ID_CONNECTOR:
        strip_from = strip_from[1:]

    if strip_from[-1] == ID_CONNECTOR:
        strip_from = strip_from[:-1]

    return strip_from


def create_agent_id(layer_name: str, prefix: str, suffix: str | None) -> str:
    """
    :param layer_name: None or the name of the layer being controlled by the agent.
    :param prefix: Prefix to prepend to the agent ID.
    :param suffix: Suffix to append to the agent ID.
    :return: Agent ID.
    """

    agent_id = f"{prefix}{ID_CONNECTOR}{layer_name}"

    if suffix is not None:
        agent_id = f"{agent_id}{ID_CONNECTOR}{suffix}"

    return agent_id


def extract_parameter_group_name(agent_id: str) -> str:
    """
    :param agent_id: Agent ID to extract parameter group name.
    :return: Extracted parameter group name.
    """

    no_prefixes = _strip_substrings(strip_from=agent_id, substrings=PREFIXES)
    layer_name = _strip_substrings(strip_from=no_prefixes, substrings=SUFFIXES)

    return layer_name


def extract_agent_type(agent_id: str) -> str:
    """
    :param agent_id: Agent ID to extract the agent type from.
    :return: Prefix designating the agent type.
    :raises ValueError: When no agent ID prefix is found.
    """

    for prefix in PREFIXES:
        if agent_id.startswith(prefix):
            return PREFIXES_TO_HPARAMS[prefix]

    raise ValueError(f"Unknown agent type with ID {agent_id}.")
